# GeoCouch 0.0.1
# Licensed under MIT.
#
# Copyright (c) 2008 Volker Mische (http://vmx.cx/)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import sys
import apsw
import simplejson as json
from couchdb.client import Server, ResourceNotFound
import pythonpath
import logging

SPATIALTE_MODULE = '/home/vmx/download/spatial/spatialite-2.2/src/versioninfo/.libs/libspatialite.so'
COUCH_SERVER = 'http://localhost:5984/'
GEODB = '/home/vmx/src/couchdb/geo/geocouch.sqlite'
LOG_FILE = '/home/vmx/src/couchdb/geo/geostore003.log'

class GeoStore(object):
    def __init__(self, geodb_name):
        logging.basicConfig(
                level=logging.DEBUG,
                filename=LOG_FILE,
                format='%(levelname) 10s %(asctime)s %(message)s')

        self.log = logging.getLogger('geostore')
        self.log.info('datebase: %s' % geodb_name)

        self.geodb_conn = apsw.Connection(geodb_name)
        self.geodb_conn.enableloadextension(True)
        self.geodb_conn.loadextension(SPATIALTE_MODULE)

        self.cursor = self.geodb_conn.cursor()
        self.couch_server = Server(COUCH_SERVER)

        # current revision of the CouchDB database. This information is used to
        # "catch up" if we want to index an already existing database
        self.seq_num = 0

        # table to store meta data about the couch tables that can't be
        # retrieved dynamically (e.g. the revision of the couchdb, the
        # last time something was added to this db)
        self.cursor.execute(
            'CREATE TABLE IF NOT EXISTS _geocouch '\
            '(db TEXT PRIMARY KEY, seq_num INTEGER);')

        self.geocouch_meta = self.get_meta()

        self.catch_up()

        self.log.debug('finished initialisation')


    def get_meta(self):
        """Meta information for spatial indexing.

        Returns a dictionary:
        key -- database name
        value -- the design document

        """
        meta = {}

        for db_name in self.couch_server:
            # database is geoenabled
            if '_design/_geocouch' in self.couch_server[db_name]:
                meta[db_name] = self.couch_server[db_name]['_design/_geocouch']

        self.log.debug('geoenabled databases: %s' % meta.keys())
        return meta

    def catch_up(self):
        """Populates the geo database if CouchDB is non-empty

        The geo database might be newly created, and the CouchDB
        already contains geo-enabled databases. This function will
        insert all documents into the geo database.

        """
        geocouch_tables = []
        for table in self.cursor.execute(
            'SELECT name FROM sqlite_master WHERE type="table";'):
            # ??? I have no idea why "table" is a tuple
            geocouch_tables.append(table[0])

        self.log.debug("Existing tables: %s" % geocouch_tables)

        for db_name in self.geocouch_meta:
            # it's already there
            if db_name in geocouch_tables:
                self.seq_num = self.cursor.execute(
                    'SELECT seq_num FROM _geocouch WHERE "db" = "%s";'
                    % db_name).next()[0]
            # not yet in geocouch (catch up)
            else:
                self.create_geodb(db_name)
                self.populate_geodb(db_name)

    def drop_geodb(self, db_name):
        """Removes a spatial database structure from the geo db"""
        self.cursor.execute('DROP TABLE %s;' % db_name);
        self.log.info('dropped a geo database (table): %s' % db_name)

    def drop_doc(self, db_name, doc_id):
        """Removes a spatial database structure from the geo db"""
        self.cursor.execute('DELETE FROM %s WHERE docid = "%s";' %
                            (db_name, doc_id));
        self.log.info('dropped document in %s: %s' % (db_name, doc_id))


    def create_geodb(self, db_name):
        """Creates a spatial database structure in the geo db"""
        self.cursor.execute('CREATE TABLE %s (docid TEXT PRIMARY KEY);'
                            % db_name)
        self.log.info('new database created: %s:' % db_name)

        meta = self.geocouch_meta[db_name]

        for geom_name in meta:
            if geom_name.startswith('_') or geom_name == 'srid':
                continue

            geom = meta[geom_name]

            # TODO support more geometry types
            if geom['type'] == 'POINT':
                # be on the safe side
                self.cursor.execute('SELECT DiscardGeometryColumn('\
                                    '"%s", "%s");' %
                                    (db_name, geom_name))
                self.cursor.execute('SELECT AddGeometryColumn('\
                                    '"%s", "%s", %i, "POINT", 2);' %
                                    (db_name, geom_name, meta['srid']))
                self.log.info('AddGeometryColumn(POINT): '\
                              'db: %s, col: %s, srid: %i' %
                              (db_name, geom_name, meta['srid']))

    def populate_geodb(self, db_name):
        """Put the geometries of all documents into the geo database"""
        couch_db = self.couch_server[db_name]

        for doc_id in couch_db:
            doc = couch_db[doc_id]

            # newly enabled (or changed) geo database
            # TODO if the "geoview" changes, recreate the
            # spatial index (geodb)
            if not doc.id.startswith('_design/'):
                self.store_geoms(doc, db_name)

        self.store_seq_num(db_name, couch_db.info()['update_seq'])


    def store_geoms(self, doc, db_name):
        """Store geometries of a specific document in the geo database"""
        meta = self.geocouch_meta[db_name]

        for geom_name in meta:
            if geom_name.startswith('_') or  geom_name == 'srid':
                continue

            geom = meta[geom_name]

            if geom['type'] == 'POINT':
                try:
                    x = pythonpath.get_item(geom['x'], doc)
                    y = pythonpath.get_item(geom['y'], doc)

                    self.store_point((x, y), geom_name, db_name, doc.id,
                                    meta['srid'])
                # not all documents have to contain the specified geo
                # information
                except KeyError:
                    self.log.debug('document (%s) doesn\'t contain a "%s".' %
                                   (doc.id, geom_name))
                    # XXX For not it's ok to delete the document, as we
                    # can can have only one geometry column. This needs
                    # to be changed when multiple columns are possible
                    # (with SpatiaLite2.3)
                    self.drop_doc(db_name, doc.id)



    def store_point(self, point, geom_name, db_name, doc_id, srid):
        """Store a point geometry in the geo database"""
        self.log.info('insert (%s) into %s: POINT %f %f' %
                      (doc_id, geom_name, point[0], point[1]))
        self.cursor.execute('REPLACE INTO %s (docid, %s)'\
                            'VALUES("%s", GeomFromText("POINT(%f %f)", %d));'
                            % (db_name, geom_name,
                               doc_id, point[0], point[1], srid))



    def store_seq_num(self, db_name, seq_num):
        self.cursor.execute('REPLACE INTO _geocouch (db, seq_num) '\
                            'VALUES ("%s", %i);' %
                            (db_name, seq_num))
        self.log.info('update sequence number in "%s": %s' %
                      (db_name, seq_num))


    def wait_for_notifications(self):
        """Main loop that is waiting for CouchDB database notifications"""
        self.log.debug('waiting for notifications')
        while 1:
            rawinput = sys.stdin.readline()

            if rawinput != '':
                self.log.debug('db notification: %s', rawinput)
                self.process_notification(json.loads(rawinput))

            else:
                self.geodb_conn.close(True)
                self.log.info('exit.')
                break


    def process_notification(self, notification):
        db_name = notification['db']

        # document is not (yet) geo-enabled enabled
        if not db_name in self.geocouch_meta and \
           '_design/_geocouch' in self.couch_server[db_name]:
            
            self.geocouch_meta[db_name] = \
                    self.couch_server[db_name]['_design/_geocouch']

        
        if db_name in self.geocouch_meta and \
           notification['type'] in ['created', 'updated', 'deleted']:

            couch_db = self.couch_server[db_name]

            # doc_props is only the document metadata, not the
            # actual document
            # XXX check if couch_db.view('_all_docs_by_seq')[seq_num:]
            # does not request *all* documents
            for doc_props in couch_db.view('_all_docs_by_seq')[self.seq_num:]:
                # doc might be deleted already
                try:
                    # ??? if couch_db[] failes, will doc contain the
                    # last working value?
                    doc = couch_db[doc_props.id]

                    # changed geodb design document
                    if doc.id == '_design/_geocouch':
                        self.log.debug('geocouch design document changed')
                        self.drop_geodb(db_name)
                        self.geocouch_meta[db_name] = doc
                        #self.create_geodb(db_name)
                        self.catch_up()
                        continue

                    self.store_geoms(doc, db_name)

                except ResourceNotFound:
                    self.log.debug('''document (%s) doesn't exist anymore''' %
                                   doc_props.id)
                    self.drop_doc(db_name, doc_props.id)

            self.seq_num = couch_db.info()['update_seq']
            self.store_seq_num(db_name, self.seq_num)


def main(args=None):
    geodb = GeoStore(GEODB)
    geodb.wait_for_notifications()


if __name__ == "__main__":
    sys.exit(main())

